#!/usr/bin/perl
###############################################################################
##                            START OF SCRIPT                                ##
###############################################################################
#
# transform plain text dictionary files to TXT and TAB format for PDIC/Zaurus
# Version 1 by Armin Rump, May 13, 2001

use Jcode;                # Initialize
use strict;

# Switches
my $doout = 1;   # 1 to activate output routines, 0 to make Zaurus data only

# Variable definitions

my %outfiles;    # $outputfiles{edict,csv7,txt} determine which output to generate (values 0 or 1)
my @tabfiles;    # records which files are being processed
my $procout = 0; # indicates whether output processing is needed

my $mainchoice = 0;
my $mainmenue;
if ($doout){$mainmenue ="\n\n
                         MAIN MENUE
                         ==========
Please type the letter of your chice, followed by the <return> key.
X   Exit
E   Explanation
I   Input file processing  (source -> tab)
O   Output file processing (tab -> desired format)
R   Reset (clear tabfile list)
";}
else {$mainmenue ="\n\n
                         MAIN MENUE
                         ==========
Please type the letter of your chice, followed by the <return> key.
X   Exit
E   Explanation
I   Input file processing (source -> one line text list)
R   Reset (clear tabfile list)
";}


my $explanation = "
Dico Brief Explanation:

SOURCE -> TABLIST -> OUTPUT FORMAT

The above graph illustrates how this script can process plain text dictionary
files from various sources to desired output formats via an internal standard
tablist format.
The easiest way to get proper source file formats is to download Jim Breen's
EDICT family of dictionaries (http://www.csse.monash.edu.au/~jwb/japanese.html)
- these files are already in plain text format.  Other sources must first be
made plain text following the instructions at http://zaurus.rump-net.de.
If you name several source files at once, they will be combined into one
dictionary file.  To create several separate dictionaries, you must go through
the input routine several times.
Output formats supported are:
TXT   - one line text lists to generate DIC files for the Sharp Zaurus PDA
CSV   - comma separed lists to use with the shareware program PDIC on a PC
EDICT - Jim Breen's format, works together with JWP, a program that runs on
      WINDOWS CE machines (http://www.physics.ucla.edu/~grosenth/japanese.html)
Most likely, you would use PDIC (http://member.nifty.ne.jp/TaN/) to further
process and use the output files.";

# Print the header
print "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";
print <<HEAD;
=============================================================================
 Perl Dictionary Conversion Program:  EDICT to tab and comma separated lists
                  Version 1, Armin Rump, March 5, 2001
=============================================================================
Intended for further processing of dictionary files, for example portation to
PDAs such as Sharp ZAURUS.
For details, see http://zaurus.rump-net.de
Note: This script uses the Perl module Jcode. Download and install Jcode before
running this script. (http://theoryx5.uwinnipeg.ca/CPAN/cpan-search.html)
-----------------------------------------------------------------------------

The general file processing scheme is as follows:
SOURCE FILES -> TAB FILES -> OUTPUT FILES

I suggest that you read the explanation the first time you run this script.
Furthermore, you may find it helpful to have a printout of its homepage
(URL above) handy for reference.

HEAD
print "Press RETURN to start";
my $tmp = <STDIN>;


###############  Start MAIN MENUE
do {
   print $mainmenue;
   $mainchoice = <STDIN>;
   $mainchoice =~ s/\n//;
   
   if ($mainchoice eq "E") {
	print $explanation;
	print "\n\n\nPress <return> to go on.";
	$tmp = <STDIN>;}
   if ($mainchoice eq "I") {
	push @tabfiles, (&inputmenue);   # right after this routine
	# subroutine queries for input files, then processes those files and 
	# returns a list of newly created tab files.  The value of $procout 
	# flag added to the list as the last element
print "\nTABFILES:\t@tabfiles";
	$procout = pop @tabfiles;
	print "\n\n\n\n\n";}
   if (($mainchoice eq "O") && ($doout) && ($procout)){
	&outputfiles(@tabfiles)}
	# queries which outputfiles to generate and converts each @tabfile
	# to the each desired format.
   if ($mainchoice eq "R"){
	@tabfiles=();
        print "\n\nTabfile List Cleared.\n\n";}
}
while ($mainchoice ne "X");                  
###############    End MAIN MENUE

if (!($doout) && ($procout)) {     # output processing if not done already
	foreach (@tabfiles) {
		   if ($_=~/\.ekh\.tab/) {&condense ($_, 0);}
		   else {&condense ($_, 2);}
	}
}                                  # endif output processing

unless ($outfiles{tab}) {          # delete undesired tab files
	foreach (@tabfiles) {unlink $_;}
	print "Temporary files deleted:  ".join ", ", @tabfiles;
}

print "\nDone: The new dictionary files are in the working directory.\n";

#####################  End Main Program dico  ##########################

sub inputmenue { # --------------------------------- sub inputmenue --
	# subroutine queries for input files, then processes those files and 
	# returns a list of newly created tab files.  The value of $procout 
	# flag added to the list as the last element
# call @tabfiles = &inputmenue;

use strict;

my $tmp;
my $procout = 0;
my $inputformats ="
  ** Jim Breen's Files (http://www.csse.monash.edu.au/~jwb/japanese.html) **
1 	EDICT
2 	ENAMDICT
3       KANJIDIC
4       LAWGLOSS
  ** Ulrich Apel's Files (http://133.1.179.17:591/WadokuJT/) **
5 	WaDokuJT
  ** EPWING CD-ROM dictionaries **
6 	Kojien (L)
  ** Others **
7 	TAB list files
";
print <<HEAD;

You may process one or more source files at once.
All source files will be combined into one dictionary.
Please name the source files you want to convert.
Type the filename followed by the return key.
Do not type space characters. Type just return when you are finished.
Type a small letter 'q' and return if you want to quit the program.

HEAD

my @sourcefiles = &inputfilenames;

print <<HEAD;

Please tell me which format template I should use to process each of
the input files.  Enter the one-digit identifiers as indicated.
After each file name, just type the proper number followed by the return
key.  (Programs in fomat 0 ["other"] will NOT be processed.)

Input file formats are:
HEAD
print $inputformats;

my @edicts;
my @wadokus;
my @tabs;
my @kojiens;
my @lawgloss;
my @enamdicts;
my @kanjidics;
my @tabdata;
my $ext;

foreach (@sourcefiles) {
  print "\n\t$_ \t";
  $tmp = <STDIN>;
  if    ($tmp=~/1/) {push @edicts,    $_; $procout++;}
  elsif ($tmp=~/2/) {push @enamdicts, $_;}
  elsif ($tmp=~/3/) {push @kanjidics, $_;}
  elsif ($tmp=~/4/) {push @lawgloss,  $_; push @edicts, $_; $procout++;}
  elsif ($tmp=~/5/) {push @wadokus,   $_; $procout++;}
  elsif ($tmp=~/6/) {push @kojiens,   $_;}
  elsif ($tmp=~/7/) {push @tabs,      $_; $procout++;}
}
print <<TTT;
OUTPUT FILE NAMES
-----------------
The tab files are generated in the working directory and will be called
*khe.tab, *hke.tab and *ekh.tab in case of the tab list format.
  k stands for kanji (or katakana), i.e. the Japanese entry
  h stands for hiragana,i.e. the pronunciation entry
  e for English, i.e. the translation entry - it may as well be German or
    whichever language dictionary you processed, the explanatory term in case of Kojien, etc.

What shall I use as the stem name of the output files (above indicated as *)?
Please chose a stem name that is unique in the current directory:
TTT
my $stemname = <STDIN>;
$stemname=~s/\n//;
& lawg2edict    (@lawgloss)             if (@lawgloss);
& edict2tab     (@edicts, $stemname)    if (@edicts);
& enam2tab      (@enamdicts, $stemname)  if (@enamdicts);
& kanjidic2tab  (@kanjidics, $stemname) if (@kanjidics);
& wadokujt2tab  (@wadokus, $stemname)   if (@wadokus);
& kojien2tab    (@kojiens, $stemname)   if (@kojiens);
foreach (@tabs) {
	($tmp, $ext) = split /\./, $_, 2;
	if (((length $ext) gt 7) && ($ext =~ /.*((e|h|k)\.tab)$/)) {
		$ext =~ s/.*((e|h|k)\.tab)$/$1/;}
	open SOURCE, "<$_";
	open TARGET, ">>$stemname.$ext";
	@tabdata = <SOURCE>;
	print TARGET @tabdata;
	close TARGET;
	close SOURCE;
	print "\nProcessed input file: $_";
}				# end foreach tabfile
my $tt1 = $stemname.".ekh.tab"; print "\n1. $tt1";
my $tt2 = $stemname.".khe.tab"; print "\n2. $tt2";
my $tt3 = $stemname.".hke.tab"; print "\n3. $tt3";
return ($tt1, $tt2, $tt3, $procout); 
} #----------------------------------------------- end sub inputmenue ---

sub lawg2edict {
my @edictpile = ();
my $line;
foreach (@_) {
    (open SOURCE, "<$_") || do {print "\nCan't read file $_ - skipped"; next;};
    while ($line = <SOURCE>) {
	$line =~s/\n//;
my	@zeile = split /\t/, $line;          # split line elements into array
	$zeile[3]=~s/^|;|$/\//g;
	$zeile[2]="[".$zeile[2]."]";
	push @edictpile, "$zeile[1] $zeile[2] $zeile[3]\n";
	}       # end while SOURCE
    close SOURCE;
    open TARGET, ">$_";
    print TARGET @edictpile;
    close TARGET;
}               # end foreach file
}               # end lawg2edict

sub outputfiles { # --------------------------------- sub outputfiles --
	# queries which outputfiles to generate and converts each @tabfile
	# to the each desired format.
# call &outpufiles(@tabfiles);
@tabfiles = @_;
print <<OUTF;
The following output formats are available:
1. TXT   - one line text lists to generate DIC files for the Sharp Zaurus PDA
2. CSV   - comma separed lists to use with the shareware program PDIC on a PC
3. EDICT - Jim Breen's format for use with JWP (for example)

Please enter all the desired numbers follwed by the <return> key.
For example, "13<return>" will generate TXT and EDICT files.
OUTF
my $outf = <STDIN>;
$outf =~s/\n//;
my @fileparts;
my $stemname;
my $endung;

if ($outf=~/1/) {             # TXT
	foreach (@tabfiles) {
		@fileparts = split /\./, $_;
		do {$endung = pop @fileparts} until ($endung =~/ekh|khe|hke/);
		$stemname = join ".", @fileparts;
		if ($stemname eq "") {print "\nfilenaming error 1 in dico.pl sub outputfiles\n";}
		if ($endung eq "ekh") {&condense ($_, 0, $stemname);}
		else {&condense ($_, 2, $stemname);}
	}
}

@fileparts = split /\./, $tabfiles[0];
do {$endung = pop @fileparts} until ($endung =~/ekh|khe|hke/);
$stemname = join ".", @fileparts;
if ($stemname eq "") {print "\nfilenaming error 2 in dico.pl sub outputfiles\n";}

if ($outf=~/2/) {             # CSV
	&tab2csv7col (@tabfiles, $stemname);	
}

if ($outf=~/3/) {             # EDICT
	&tab2edict (@tabfiles, $stemname);	
}

} #----------------------------------------------- end sub outputfiles ---

sub edict2tab { #  ***********************************************************
#                  This subroutine extracts edict files and
#                  puts it into separate files for kanji, hiragana and
#                  English keywords, respectively.  The output files are tab-
#                  separated lists.
#                  ***********************************************************
# open output file handles
my $stem = pop @_;
open KHE, ">$stem.khe.tab";
open HKE, ">$stem.hke.tab";
open EKH, ">$stem.ekh.tab";

# start working on input files
foreach (@_) {
    (open EDICT, "<$_") || do {print "\nCan't read file $_ - skipped"; next;};
my  $line;
    while ($line = <EDICT>) {                        # processing file entry
        $line = jcode($line)->euc;           #convert to EUC-JP just to be sure
	$line =~s/\/\((P|(0|1|2|3|4|5|6|7|8|9|-)*)\)\//\//;
                                             #remove nonsense parenthesis terms
	$line =~s/~|||_//g;              # remove gargabe characters
my	$kanji = "";
my	$hira;
	if ($line =~ /\[.*\] /){$kanji = "1";}#$kanji set if kanji entry exists
	$line =~ s/\t/ /g;                   # replace tabs by spaces to
	                                     # eliminate bad separators
	$line =~ s/ \[|\] \/| \/|\//\t/g;    # make line tab-separated
	$line =~ s/\n//;                     # and delete line endings
my	@zeile = split /\t/, $line;          # split line elements into array
	foreach (@zeile) {$_=~s/^\s*|\s*$|\n//;}   #discard whitespace
	if ($kanji) {
		$kanji= shift @zeile; # kanji entry taken out of zeile to $kanj
		$hira = shift @zeile; # hiragana likewise
	    }
	    else {$hira = shift @zeile;}       # @zeile retains only English.
	#  -------------------------------     write to files khe and hke
	$"="; ";                      # set array element separator
	print KHE "$kanji\t$hira\t@zeile\n";
	print HKE "$hira\t$kanji\t@zeile\n";
	$"="";                        # and reset the separator

	#  -------------------------------     process the English entries

my	$current = shift @zeile;

	# the first English entry often has a prefix in paraenthesis
	# which must be removed.  "(phys.) quantum mechanics" is an example.

my	($prefix, $garbage, $current)=$current=~/^\s*((\(.+?\)\s*)*)\s*(.*$)/;

	#      regexp. assignment in words:
	# at beginning ^ of string $current skip any whitespace \s there may 
	# be *, then assign to $prefix () as many repeats * of the following
	# expression as possible: ((assign to $garbage))
	# as many characters . as necessary ? (but at least one) between
	# parentheses \( \) (must be escaped as special characters).
	# then skip additonal white space and assign to $current () the 
	# remaining characters .* up to the end of the string $.

          do {                           # print each English entry
	   print EKH "$current$prefix\t$kanji\t$hira\n";
	    $current = shift @zeile;
	} while ($current =~ /\w/);        # as long as there is an 
                                           # alphanumeric English entry
     }                   # end while ($line = <edict>)
   close EDICT;
   }                     # end foreach (@ARGV): finished processing file entry
# close output filehandles
close EKH;
close HKE;
close KHE;
print "\ntab files generated: ekh.tab, hke.tab, hke.tab\n";
}                        # end sub edict2tab

#  ================================================  Sub Inputfilenames ===
sub inputfilenames {
my @gotit;
my $input = 1;
while ($input) {
  $input = "";
  print "Enter: ";
  $input = <STDIN>;
  $input =~ s/\n|\s//g;               # get rid of whitespace
  if ($input eq "q"){die "Program aborted by user request";}
  if ($input && (! -f $input)){       # check if input is valid file name
        print "File $input does not exist.  Please enter again.\n";
        redo;
  }
  push @gotit, $input;
}
pop @gotit;         # discard the last entry (empty return)
return @gotit;
}
sub enam2tab { #  ***********************************************************
#                  This subroutine extracts edict files and
#                  puts it into separate files for kanji, hiragana and
#                  English keywords, respectively.  The output files are tab-
#                  separated lists.
#                  ***********************************************************
# open output file handles
# Engl. output format:  "english term (suffix1;suffix2)"

my $filestem = pop;
my $tabkh = ">".$filestem.".kh.tab";
my $tabhk = ">".$filestem.".hk.tab";
my $onefile = ">".$filestem.".txt";
open KH, $tabkh;
open HK, $tabhk;
open TXT, $onefile;

# start working on input files
foreach (@_) {
    print "\nprocessing $_ .. ";
    (open EDICT, "<$_") || do {print "\nCan't read file $_ - skipped"; next;};
my  $line;
    while ($line = <EDICT>) {                        # processing file entry
        $line = jcode($line)->euc;           #convert to EUC-JP just to be sure
my	$kanji = 0;
	if ($line =~ /\[.*\] /){$kanji = "1";}#$kanji set if kanji entry exists
	$line =~ s/\t/ /g;                   # replace tabs by spaces to
	                                     # eliminate bad separators
	$line =~ s/ \[|\] \/| \/|\//\t/g;    # make line tab-separated
my	$addengl = "";
	if (($addengl) = $line=~/\((\w(\w| )+)\)/) {
		if ($addengl =~/\d/) {$addengl = "";} }
my	%fill;
	$fill{"p"}="place";
	$fill{"h"}="historical person";
	$fill{"s"}="surname";	
	$fill{"u"}="person";
	$fill{"g"}="given name";
	$fill{"m"}="male";
	$fill{"f"}="female";
	$line =~ s/\((\w)(\)|,)/($fill{$1}$2/;  # expand abbreviations
	$line =~ s/, ?(\w),/, $fill{$1},/g; 
	$line =~ s/, ?(\w)\)/, $fill{$1})/; 
	$line =~ s/\n//;                     # and delete line endings
my	($shortline, $suffix) = $line =~/(.*?)(\(.*\))/;
	$suffix =~ s/\t/; /g;
my	@zeile = split /\t/, $shortline;     # split line elements into array
	if ($kanji) {                        # Print: kan->hira(), hira->kan()
my	  $test = $zeile[0];
	  $test =~ s///;
	  $test =~ s///g;
	  unless ($test eq $zeile[1]) {
	    print KH "$zeile[0]\t$zeile[1]\t$suffix\n";
	    print HK "$zeile[1]\t$zeile[0]\t$suffix\n";
	    print TXT jcode("$zeile[0] /// $zeile[1] $suffix\n")->sjis;
	    print TXT jcode("$zeile[1] /// $zeile[0] $suffix\n")->sjis;
		
	 }
	 if ($addengl) {		     # English names with Kanji
	    $suffix =~s/\($addengl\)//;
	    if ($test eq $zeile[1]) {        # kanji redundant
		print HK lc($addengl)."\t$zeile[0]\t$suffix\n";
		print TXT jcode(lc$addengl." /// $zeile[0] $suffix\n")->sjis;
		}
	    else {
		print HK lc($addengl)."\t$zeile[0] ($zeile[1])\t$suffix\n";
		print TXT jcode(lc($addengl)
		            ." /// $zeile[0] ($zeile[1]) $suffix\n")->sjis;
	     }
	  }			       # endif addengl
	}                              # endif kanji
	else {			       # Print: kat->engl(), engl->kat()
my	    $hira = &kata2hira($zeile[0]);
	    if ($hira ne $zeile[0]) {  # if there are any katakana
my		$kata = shift @zeile;
my		$engl = join "; ", @zeile;
		if ($addengl) {
			$suffix =~s/$addengl//;
			$suffix =~s/\(\)|^ *| *$//g;
		}
		$engl =~s/\(\)| *$//g;
		$engl =~ tr/-/h/;
		print KH "$kata\t$engl\t$suffix\n";
		print TXT jcode("$kata /// $engl $suffix\n")->sjis;
		print HK lc($engl)."\t$kata\t$suffix\n";
		print TXT jcode(lc($engl)." /// $kata $suffix\n")->sjis;
		if ($addengl) {
			print TXT jcode(lc($addengl)
			            ." /// $kata $suffix\n")->sjis;
			print HK lc($addengl)."\t$kata\t$suffix\n";
			}
	     }		  # endif entry not redundant
	}                 # endelse print katakana
      }                   # end while ($line = <edict>)
   close EDICT;
   print "Done.";
   }                     # end foreach (@ARGV): finished processing file entry
# close output filehandles
close KH;
close HK;
close TXT;
}                        # end sub edict2tab

#--------------------------------------  kata2hira
# call &kata2hira ($katakanastring);
# returns the string converted to hiragana
sub kata2hira
{
my %hira;
my $line;
my @khr;
my $result;
my $start;
$start = $result = shift @_;

open CONV, "<kanaconv.tab";
while ($line = <CONV>) {		# create conversion hash
	@khr = split "\t", $line;
	$hira{$khr[1]} = $khr[0];
}
close CONV;

foreach (keys%hira) {$result =~ s/$_/$hira{$_}/g; }
return $result;
} 				# end of sub kata2hira
sub wadokujt2tab { #  ***********************************************************
#                  This subroutine extracts edict files and
#                  puts it into separate files for kanji, hiragana and
#                  English keywords, respectively.  The output files are tab-
#                  separated lists.
#                  ***********************************************************
# open output file handles
my 	$i;
my	$kanji = "";
my      @kanjis = ();
my	@cleankanjis = ();
my	$hira;
my      $cleanhira;
my	$deutsch;
my	@dtlist = ();
my	@cleandtlist = ();
my  $dtmp;
my  $line;
my  @zeile;
my $stem;
my $maxlength = 50; #max. lenth of German entries (used to filter out examples)
my $word;
my $pfix;
my @worte;
my @tt;
my $tmp;
$stem = pop @_;
open KHE, ">>$stem.khe.tab";
open HKE, ">>$stem.hke.tab";
open EKH, ">>$stem.ekh.tab";

# start working on input files
print "\nReading in files: ";
foreach (@_) {
    (open EDICT, "<$_") || do {print "\nCan't read file $_ - skipped"; next;};
    print "$_ .. ";
    my $dumpfirstline;
    print $dumpfirstline = <EDICT>;
    while ($line = <EDICT>) {                        # processing file entry
# $i++;
# (($i%1000) || (print "\n$i\t$line"));
        $line = jcode($line)->euc;           #convert to EUC-JP just to be sure
	$line =~ s/\n//;
	@zeile = ();
	@zeile = split /\t/, $line;          # split line elements into array
	foreach (@zeile) {$_=~s/^\s*|\s*\n?$//;}   #discard whitespace
	$kanji= shift @zeile; # kanji entry taken out of zeile to $kanj
	$cleanhira = $hira = shift @zeile; # hiragana likewise
       	$deutsch = join "; ", @zeile;	     # @zeile retains only German.
	$cleanhira =~ s/\[\d+\]//g;               # clean up hira
	$cleanhira =~ s/^ +|| +$//g;
	@kanjis = split /;|\(|\)/, $kanji;
	@cleankanjis = ();
	foreach (@kanjis) {
		$_ =~ s/\[\d+\]//g;               # clean up kanjis
		$_ =~ s/^ +|| +$//g;
		if (($_!~/$cleanhira/)&&($_=~/\W\W/)){push @cleankanjis, 
			(jcode($_)->sjis);}
	}					  # endforeach kanji
	$deutsch =~ s/\. \[/; \[/g;
	$deutsch =~ s/\((.*); (.*)\)/($1, $2)/g;
	$deutsch =~ s/\{(.*); (.*)\}/($1, $2)/g;
	@cleandtlist = ();
	@dtlist = split /;|\/\//, $deutsch;
	foreach (@dtlist) {
            if ((length $_ < $maxlength) && ($_!~/\{Bsp\.\}/)) { 
            					# clean up deutsch
		$_ =~ s/\. \[/; \[/g;	# . [	; [
		$_ =~ s/\[\d+\] *//g;     #[..]	delete          
		$_ =~ s/\.?\n//;	#.ret	delete
		$_ =~ s/^\.+//;		#"
		$_ =~ s/\.\s*$//g;	#"
		$_ =~ s/\{(.*)\} *(.*) *$/$2 \($1\)/;	#{..}	postfix
		$_ =~ s/^ *(\(.*?\) *)*(.*) *$/$2 $1/;#(..)	postfix
		$_ =~ s/^(\W|\d)*//;			#nonalphab postfix
		if ($_=~/^\w/){push @cleandtlist, (lc $_);}
		if ($_=~/\w ?\.\.\. ?\w/) {    # postfix stuff preceeding ...
			$_=~s/^(\w(\w| )+)\.\.\. ?(\w[^\(]*)/$3, $1/;
			push @cleandtlist, (lc $_);}
	$dtmp = $_;
	if($dtmp=~s/^(sich) (\w[^\(]*)/$2, $1/){push @cleandtlist, (lc $dtmp);}
	if($dtmp=~s/^(eine?n?) (\w[^\(]*)/$2, $1/)
		{push @cleandtlist, (lc $dtmp);}
	if($dtmp=~s/^((d|w)(e|i|a)(n|e|r|s|m)) (\w[^\(]*)/$5, $1/)
		{push @cleandtlist, (lc $dtmp);}
	if($dtmp=~s/^(i(n|m)) (\w[^\(]*)/$3, $1/)
		{push @cleandtlist, (lc $dtmp);}
	if ($_ =~ /\)$/) { ($word, $pfix) = $a =~ /^(.*?)(\(.*\))$/;}
	 else {$word = $_; $pfix = "";}
	@worte = ();
	if (($word =~ / /)&&(ord $word > 96)&&(ord $dtmp > 96)) {
	   if ($word =~ /\w$/) {$word = $word.",";}
	   @tt = split " ", $word;
	   while (@tt > 0) {
		push @worte, (shift @tt);
		if ((ord $tt[0]<91)&&(ord $tt[0]>64)) {last;}
	   }
	if (@tt){push @cleandtlist, (lc join " ", (@tt,@worte,$pfix));}
}  # endif a has space

	    }					 # endif not an example phrase
	}                                        # end foreach dtlist
	if ($kanji =~ /$cleanhira/) {$tmp = 1} else {$tmp = 0}
        $kanji = jcode($kanji)->sjis;        #convert back to Shift-JIS
        $hira  = jcode($hira) ->sjis;        
        if ($cleanhira =~/\w/) {$cleanhira = ""}
         else { $cleanhira = jcode($cleanhira)->sjis;}          
	#  -------------------------------     write to files
	if (@cleankanjis) {if ($cleanhira) {
	    print HKE"$cleanhira\t".(join ", ", @cleankanjis)."\t$deutsch\n";
	}}
	else {if ($cleanhira) {print HKE"$cleanhira\t\t$deutsch\n";}}
	if ($tmp) {
		foreach (@cleandtlist){ 
		    print EKH "$_\t".(join ", ", @cleankanjis)."\t$hira\n";}}
	  else {foreach (@cleandtlist) {
			print EKH "$_\t".$kanji."\t$hira\n";}} 
	foreach (@cleankanjis) { print KHE "$_\t$hira\t$deutsch\n";}
     }                   # end while ($line = <edict>)
   close EDICT;
   }                     # end foreach (@ARGV): finished processing file entry
# close output filehandles
close EKH;
close HKE;
close KHE;
print "\nWadokuJT tab files generated: $stem.ekh.tab, $stem.hke.tab, $stem.khe.tab\n";

}                        # end sub edict2tab

sub kanjidic2tab {#***********************************************************
#                  This subroutine extracts edict files and
#                  puts it into separate files for kanji, hiragana and
#                  English keywords, respectively.  The output files are tab-
#                  separated lists.
#                  ***********************************************************
# open output file handles
my $line;
my @term;
my $term;
my $id;
my $content;
my $kanji;
my @china;
my @korea;
my $freq;
my $pfreq;
my $engl;
my @english;
my @yomi;
my @henyomi;
my $phenyomi;
my @addyomi;
my $tmpyomi;
my $ttyomi;
my $nanori;
my $i = 0;

my $stem = pop @_;
open KHE, ">>$stem.khe.tab";
open HKE, ">>$stem.hke.tab";
open EKH, ">>$stem.ekh.tab";
open KDIC,">kanjidic.txt";

# start working on input files
print "\nReading in files: ";
foreach (@_) {
    (open EDICT, "<$_") || do {print "\nCan't read file $_ - skipped"; next;};
    print "$_ .. ";
    while ($line = <EDICT>) {  		# processing file entry
        if ($line =~s/^# //) {          # prefix line
#        	print KHE "0\t$line";   # print Jim Breen's reference
#		print KDIC "0 /// $line";
		next;}
	@china =();			# reset the variables
	@korea =();
	$freq = "";
	$engl;
	@yomi = ();
	@henyomi = ();
	@addyomi = ();
	$nanori = 0;
	
        $line = jcode($line)->euc;   	# convert to EUC-JP just to be sure
	$line =~ s/\n//;
	@term = split " ", $line;		# define variables
	$kanji = jcode(shift @term)->sjis;
	while (@term) {
	   $term = shift @term;
	   $id = "";
	   $content = "";
	   if ($term =~/\w/) {($id, $content) = $term =~ /(.)(.*)/; }
	   else {$id = "J"; $content = $term;}
	   $nanori = 1 if ($id eq "T");
	   
	   $freq = "grade $content " if ($id eq "G");
	   $freq = $freq."freq. $content"if ($id eq "F");
	   push @china, ($content)   if ($id eq "Y");
	   push @korea, ($content)   if ($id eq "W");
	   if ($id eq "J") {
		if ($nanori) {push @henyomi, (jcode($content)->sjis);}
		else {	
		   push    @yomi, (jcode($content)->sjis);
		   unless ($content =~ /\./) {
			 push @addyomi, (jcode($content)->sjis);
			$ttyomi = &kata2hira($content);
			push @addyomi,(jcode($ttyomi)->sjis) 
				if($ttyomi lt $content);
		   }
		   else {
			$tmpyomi = $content;
			$content =~ s/\..*//;
			$ttyomi = &kata2hira($content);
			push @addyomi, (jcode($ttyomi)->sjis) 
				unless ($ttyomi eq $content);
			push @addyomi, (jcode($content)->sjis);

			$tmpyomi =~ s/\.//;
			$ttyomi = &kata2hira($tmpyomi);
			push @addyomi, (jcode($ttyomi)->sjis)
				unless ($ttyomi eq $tmpyomi);
			push @addyomi, (jcode($tmpyomi)->sjis);
		   }			# end else point
		}			# endelse nanori
	   }				# end J
	   if ($id !~/\w/) {unshift @term, ($content); last;}
	}				# end foreach @term

	$engl = join " ", @term;        # @term retains only English terms
	$engl =~ s/\} \{/, /g;
	$engl =~ s/\{|\}//g;

	foreach (@china) {
		$_=~s/1/-/;		# Chinese pronunciation symbols;
		$_=~s/2/\//;
		$_=~s/3/\^/;
		$_=~s/4/./;
		$_=~s/5/\=/;}
	foreach (@yomi) {$_ = jcode($_)->sjis;}
	
	# print tab files
	if (@henyomi) {$phenyomi = " (".(join ", ", @henyomi).")";}
	else {$phenyomi = "";}
	if ($freq=~/\d/) {$pfreq =" <$freq>";}
	else {$pfreq = "";}
	print KHE "$kanji\t".(join ", ", @yomi)."$phenyomi\t$engl$pfreq [:"
		.(join ", ", @china)." :".(join ", ", @korea)."]\n";
	foreach (split ", ", $engl) {print EKH "$_\t$kanji\n";}
	foreach (@addyomi, @henyomi) {
		print HKE "$_\t$kanji\n"; }
	
	# print onetext file
	 print KDIC "$kanji /// ".(join ", ", @yomi).
	 	"$phenyomi \\ $engl \\ $freq  :"
	 	.(join ", ", @china)." :".(join ", ", @korea)."\n";
	foreach (split ", ", $engl) {print KDIC "$_ /// $kanji\n";}
	foreach (@addyomi, @henyomi) { print KDIC "$_ /// $kanji\n"; }
    } 				# end while edict
    close EDICT;
    print "Done.\n";
}				# end foreach source file
close KHE;
close HKE;
close EKH;
close KDIC;
}				# end of subroutine kanjidic2tab

# ---------------------------------------------  kata2hira  ------------
# call &kata2hira ($katakanastring);
# returns the string converted to hiragana
sub kata2hira
{
my %hira;
my $line;
my @khr;
my $result;
my $start;
$start = $result = shift @_;

open CONV, "<kanaconv.tab";
while ($line = <CONV>) {		# create conversion hash
	@khr = split "\t", $line;
	$hira{$khr[1]} = $khr[0];
}
close CONV;

foreach (keys%hira) {$result =~ s/$_/$hira{$_}/g; }
return $result;
} 				# end of sub kata2hira
sub koji2tab {
my $line;
my $infile;
my $outfile;
my $active = 0;
my $last = 0;
my @entry = ();
my $setsumei;
my @eintrag = ();
my $tab = "\t";
my $return = "\n";
my $hira;
my $tempkanji;
my $kanji;
my @kanjis;
my $i = 0;
my $pointer = " ";
my $kline = "";
my $csvline;
my $eline;
my $tabfile;
my $onelinefile;
my $ji;
my $okuri;
my $altern;


# ----------  GAIJIDEL HASH --
my %gaiji;
my $preprohash = "";
my @prohash= ();
# create the replacement hash:
open GAIJIFILE, "<kojien.gaiji";
# hashfile format: key<tab>value<tab>comment
while ($preprohash = <GAIJIFILE>) {
   @prohash = split /\t/, $preprohash;
   $gaiji{$prohash[0]} = $prohash[1];
}                   		         # end while gaijiline
close GAIJIFILE;


my $file = $_[0];
$file =~ s/\..*//;
$infile = $file.".txt";
$tabfile= $file.".tab";
$onelinefile= $file."one.txt";
open SOURCE, "<$infile";
open TABOUT, ">$tabfile";
open ONEOUT, ">$onelinefile";

# do {$line = <SOURCE>} until ($line eq '\n');    # fastforward through intro
# if(/^$/../Ȥֽʡ$/)               # alternative
push @entry, ($line);
print "\nCreating Kojien tab file:  $outfile ...";

while ($line = <SOURCE>) {
    if ($line eq "\n") {                # produce an entry on empty line
      if (($entry[0] =~ /^[^\s]{2}/) && ($entry[1] =~ /^  [^\s]{8}/)) {
	my $symA = "";
	my $symB = "";
	($hira, $kanji) = split /$symA/, shift @entry; # process the definition
	$hira =~ s/|\n//g; 
	($ji, $okuri)   = split /$symB/, $hira;
	$hira =~ s///g;                           # clean up
	$kanji =~ s/.*?\n//;
	@kanjis = split /$symB/, $kanji;
	if ($okuri) {
		$ji = $kanjis[0];
		$ji =~ s/$okuri$//; 
	}
	foreach $eline (@entry) {		# clen up the explanation
		$eline =~ s/\n//;                
		$eline =~ s/\[(z\w{4})\]/$gaiji{$1}/g;      # GAIJIDEL
		$eline =~ s/^\s*//;
		$eline =~ s/.*?//g;
		$eline =~ s/.*?$//;
		}
	 $setsumei = join "\t", @entry;
	 $setsumei =~ s//$ji/g;
	 $setsumei =~ s//$ji/g;
	 $setsumei =~ s/$kline/$kanjis[0]/g;
	if (($kanji) && !($kanji =~ /\w/ )) {    # unless gairaigo or hira only
	  $i = @kanjis;
	  @eintrag = ();
	  $altern = "";
	  if ($i > 1) {$i = 0;
	     $altern = '  (';
	     foreach (@kanjis) {
		$eintrag[$i] = $_.$tab.$pointer.$kanjis[0].$return;
		$altern = $altern.$_."/" if ($i > 0);
		$i++;
		}
	     $altern =~ s/\/$/)/;
	  }                          # end if altern. kanjis
	  $eintrag[$i] = $hira.$tab.$pointer.$kanjis[0].$return;
	  if ($altern =~ /\(.+\)/) {
		$eintrag[0] = $kanjis[0].$tab.$hira.$tab.$altern.
				$tab.$setsumei.$return; }
	   else {$eintrag[0] = $kanjis[0].$tab.$hira.$tab.$setsumei.$return; }
	}				# end if no kanji
	else {$eintrag[0] = $hira.$tab.$kanji.$tab.$setsumei.$return;}
					# in this case, $kanji is empty or is
					# the gairaigo origin (often English)
	foreach (@eintrag) {            # print the output files
		$_ = jcode($_)->sjis, "euc"; # back so Shift-JIS
		print TABOUT $_; 
		$_ =~ s/\t/ \/\/\/ /;
		$_ =~ s/\t\(([^\t]+)\)\t/  $1 \\ /;  # add alt. kanji to hira
		$_=~ s/\t/ \\ /g;
		print ONEOUT $_;
	}
      }                                 # endif valid entry
      @entry = ();			# reset entry
      @eintrag = ();
    }                                   # end if line empty
   else {                               # build entry from lines
        $line = jcode($line)->euc, "sjis";       # convert EUC-JP to Shift-JIS
	push @entry, ($line);
# print "Line read: \t $line";
	}
 }					# end while SOURCE
 close TABOUT;
 close ONEOUT;
 close SOURCE;
 print " Done.";
}			# end sub koji2tab
sub condense {
my @nextone;
my $suf;
my $file = $_[0];
my ($stem) = $file=~/\.((h|k|e){3})\b/;
    $stem = $_[2].".".$stem;
my $filename = "<$file";
my $suffixpos = $_[1];
if (open (SOURCE, $filename)) { 
 my @dico = <SOURCE>;
 close SOURCE;
 @dico = sort @dico;
 my @condense = ();
 my @current;
 my $hit;  
 my $tail;
 do { @current = &cleanentry (jcode(pop @dico)->euc);   }
   until (($current[$suffixpos]=~/^(\(\w*\) ?)?\w/) && ($current[$suffixpos]=~/^\D/) 
     && ($current[2-$suffixpos]=~/\S/));
   my @csuffixlist = &getsuffix ($current[$suffixpos]);
 if($current[1]) {$current[2]= "$current[1] - $current[2]";
	$current[1]="";}              # e->k-h, k->h-e, h->k-e
 print "\nCondensing file: $filename  ...  ";
 while (@dico) {                  # deconstruct dico (and construct condense)
  @nextone = &cleanentry (jcode(pop @dico)->euc);
  foreach (@nextone) {
	$_=~s/\[|\]|\\|\*|\+|\?| \d\) //g;            # clean up the terms
	$_=~s/ *\([^\)]*($|\/\/)//;   # remove erroneous parentheses
	$_=~s/(^|\/\/)[^\(]*\) *//;
	}
  my @nsuffixlist = &getsuffix ($nextone[$suffixpos]);
  if (($nextone[$suffixpos]=~/^(\(\w*\) ?)?\w/) && ($nextone[$suffixpos]=~/^\D/)     && ($nextone[2-$suffixpos]=~/\S/)) {                  # if valid entry
     if ($current[0] eq $nextone[0]) {      # <0> are same, combine them
	if ($suffixpos == 0) {@csuffixlist = 
		&combinelists (@csuffixlist, @nsuffixlist);}
#	if ($current[2]=~/$nextone[1](,| )/) {             # if <1> are same
#	  unless ($current[2] =~/ $nextone[2](,| |$)/) {  
	if ($current[2]=~/$nextone[1]/) {             # if <1> are same
	  unless ($current[2] =~/ $nextone[2]/) {  
	    $hit = (index $current[2], $nextone[1]);
	    $current[2] =~ s/^(.{$hit}.*? - )/$1$nextone[2], /; 
	    if ($suffixpos == 2) {@csuffixlist = 
		&combinelists (@csuffixlist, @nsuffixlist);}
          }
	}
	else {                              		 # if <1> are different
	  if ($current[2] =~/ $nextone[2](,| |$)/) { 
	     $hit = (index $current[2], $nextone[2]);
	     $tail = length $current[2] - $hit;
	     $current[2] =~ s/ - (.*?.{$tail})$/, $nextone[1] - $1/;
	       if ($suffixpos == 0) {@csuffixlist = 
		&combinelists (@csuffixlist, @nsuffixlist);}
	  }
	  else {                                  # <1> and <2> both different
	     if (($suffixpos == 2) && (@csuffixlist[0]=~/\w/)) {
	       $suf = join /; /, @csuffixlist;
	       $suf = " ($suf)";}
	     elsif ($suffixpos == 0) 
		{@csuffixlist = &combinelists(@csuffixlist, @nsuffixlist);
		$suf = "";}
	     else {$suf = "";}
	     $current[2] = $current[2].$suf." \\ $nextone[1] - $nextone[2]";
	     @csuffixlist = @nsuffixlist;
	  }
	}                                                #endelse <1> different
	@nextone = @current;
	@nsuffixlist = @csuffixlist;
	}
     else {                                              # <0> different
	$suf = join /; /, @csuffixlist;
	if ($suf =~ /\w/) {$suf = " ($suf)";} else {$suf = "";}
	$current[$suffixpos] = $current[$suffixpos].$suf;
	unshift @condense,(lc $current[0]." /// $current[2]");
	}
  @current=@nextone;          # switch variables for next iteration step
  if($current[1])  {$current[2] = "$current[1] - $current[2]";
	$current[1] = "";}
  }                           # endif meaningful entry
}                             # endwhile (@dico)


# record last entry in case it was skipped (<0> equal)
if (($current[$suffixpos]=~/^\w/) && ($current[$suffixpos]=~/^\D/) 
     && ($current[2-$suffixpos]=~/\S/)) {
  unshift @condense,(lc $current[0]." /// $current[2]"); }

$filename = ">".$stem.".txt";
open (TARGET, $filename);
print TARGET "0 /// Dictionary file genearted by Armin Rump's script various 
dictionary files (eg. EDICT by Jim Breen, WadokuJT by Ulrich Apel, EPWING format files, etc.) \\ Details at: http://zaurus.rump-net.de\n";
foreach (@condense) {  print TARGET (jcode ("$_\n")->sjis);}
close TARGET;
print "$filename. ";
}}                              # endif fileopen, end of sub condenseekh


sub cleanentry { # ----------------------------  sub cleanentry ------
my @entry;
my $pr;
my $tr;
  @entry = split /\t/, $_[0];
  # clean up engl. entry (each shall start with lower case letter)
        foreach (@entry) {$_=~s/^\s*|\s*\n//g;}            # white spaces
	if (($pr, $tr) = $entry[0]=~ /^(\(.+?\)) (\w.*)$/) # parenthesis
	   {if ($tr =~ /\($pr\)/) {$entry[0] = $tr;}
	    else {$entry[0]=$tr.$pr;}
	   }                                                 # endif parenth.w
	$entry[0]=~ s/^(\(.+?\)) (\w.*)$/\2\1/; # parenthesis
	$entry[0]=~ s/"//g;                     # quotes
	$entry[0]=~ s/^-\s*//;
	lc $entry[0];                           # lower case
	$entry[0]=~ s/^(an?) (\w[^(\(|\))]*\w)/$2, $1/;      # preceeding 'a'
        $entry[0]=~ s/^an? (\((\w| )*?\))\s*(\w(\w| )*\w)/\3, a \1/;
	$entry[0]=~ s/^to (\w[^(\(|\))]*\w)/\1, to/;         # preceeding 'to'
@entry;                              # return value
}                                    # endsub cleanentryekh

sub getsuffix {  #  -----------------------  getsuffix ----------------
# call &getsuffix ($suffixedstring)
# returns: @suffixlist
my $sterm;
my @slist=();
while ($sterm = $_ =~ /^ *\([^(\(|\))]*\)|\([^(\(|\))]*\) */) {
  $_=~s/$sterm|^ *| *$//g;
  $sterm =~ s/^ *\(|\) *$//g;
  push @slist, ($sterm);
  }
return @slist;
}

sub combinelists {   # --------------------- combinelists --------------
# call &combinelists (@list1, @list2, .. @listn)
# returns: a concatenated list with identical elements eliminated
my @newlist = (shift @_);
while (@_) {push @newlist, shift @_;}
return @newlist;
}                # endsub combinelists
sub tab2csv7col   # =========================== tab2csv7col  =============
# 7column csv files are suitable for use with PDICT under Windows.
# column 1- def, 2- transl, 3- example, (4level 5flag 6flag) 7- pronunciation
# input files are 3 column tab
# CALL: $tab2csv7col (@tabfiles, $stemname);
{
my $stemname = pop;
my $version = 0;
my $filename;
foreach (@_){
  my  @filepile = split /\./, $_;
  pop @filepile if (@filepile > 2);
  $filepile[0]=$stemname;
  $filename = join ".", (@filepile, "csv");
  while (-f $filename) {
    $version++;
    $filepile[0]=$stemname.$version;
    $filename = join ".", (@filepile, "csv");}
  open SOURCE, "<$_";
  open TARGET, ">$filename";
  my $line;
  my $div = '","';
  my $quote= '"';
  my $link = '","",0,0,0,"';
  while ($line = <SOURCE>) {
        $line = jcode($line)->sjis, "euc";        # convert EUC-JP to Shift-JIS
	$line =~ s/\n//;                          # discard newline characters
	$line =~ s/\B"(\w(\w|\s)*?\w)"\B/<$1>/g;  # replace quotation marks
	$line =~ tr/"/;/;
     my @entry = split /\t/, $line;
     if ($filename =~/\.ekh\./){   # invert ekh to ehk
	my $hira  = pop @entry;
	my $kanji = pop @entry;
	push @entry, ($hira, $kanji);
	}
     my $seven = $quote.$entry[0].$div.$entry[1].$link.$entry[2].$quote."\n";
     print TARGET $seven;
   }                  # end of while line 
  close TARGET;
  close SOURCE;
}		      # end of foreach file
}                     # end of sub tab2csv7col

# =================================== sub tab2edict ======================
# subroutine tab2edict
# call &tab2edict (@tabfiles, $stemname)
# produces edict files from 3column tab list dictionaries
sub tab2edict {
my $stemname = pop;
my $version = 0;
my $filename;
my $line;

foreach (@_){
  my  @filepile = split /\./, $_;
  pop @filepile if (@filepile > 2);
  $filepile[0]=$stemname;
  $filename = join ".", (@filepile, "edict");
  while (-f $filename) {
    $version++;
    $filepile[0]=$stemname.$version;
    $filename = join ".", (@filepile, "edict");}
  open SOURCE, "<$_";
  open TARGET, ">$filename";
  while ($line = <SOURCE>) {
#        $line = jcode($line)->sjis, "euc";       # convert EUC-JP to Shift-JIS
	$line =~ s/\n//;                          # discard newline characters
	$line =~ s/\/|\[|\]/-/g;  # eliminate inappropriate divider marks
     my @entry = split /\t/, $line;
     my $edict = $entry[0]."[".$entry[1]."] /".$entry[2]."/\n";
     print TARGET $edict;
     }                  # end of while line 
  close TARGET;
  close SOURCE;
}		      # end of foreach file
}                     # end of sub tab2edict
